
/*
 * Text console device (console.c)
 * Part of libpogo, a c-library replacent for GBA
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 * Device for text output. Supports fixed and proportional fonts,
 * multiple windows etc.
 *
 **/


/*
TODO

- Resizing fixed windows does not update lines or columns. Buffer must be reallocated/

*/


#include "pogo/stdcore.h"
#include "pogo/device.h"
#include "pogo/font.h"
#include "pogo/screen.h"
#include "pogo/console.h"

//#include <assert.h>

#define assert(x) (void)(x)

#ifdef DS
#define TO_RGB16(r,g,b) ( ((b << 7) & 0x7C00) | ((g << 2) & 0x03E0) | (r >> 3) )
#else
#define TO_RGB16(r,g,b) ( ((r << 8) & 0xF800) | ((g << 3) & 0x07E0) | (b >> 3) )
#endif

#define TO_RGB32(r,g,b) ((r<<16) | (g<<8) | b)


//#define UINT2PIXEL(x) ( (uint16)( ((x & 0xF80000)>>8) | ((x & 0xFC00)>>5)  | ((x & 0xF8)>>3) ) )

#define FLUSH() console_putchar(0);
/*
const static uint16 ansi_colors[16*3] =
{
	TO_RGB16(0x00,0x00,0x00),
	TO_RGB16(0xC0,0x00,0x00),
	TO_RGB16(0x00,0xC0,0x00),
	TO_RGB16(0xC0,0xC0,0x00),
	TO_RGB16(0x00,0x00,0xC0),
	TO_RGB16(0xC0,0x00,0xC0),
	TO_RGB16(0x00,0xC0,0xC0),
	TO_RGB16(0xC0,0xC0,0xC0),

	TO_RGB16(0xC0,0x70,0x30),
	TO_RGB16(0xFF,0x00,0x00),
	TO_RGB16(0x00,0xFF,0x00),
	TO_RGB16(0xFF,0xFF,0x00),
	TO_RGB16(0x00,0x00,0xFF),
	TO_RGB16(0xFF,0x00,0xFF),
	TO_RGB16(0x00,0xFF,0xFF),
	TO_RGB16(0xFF,0xFF,0xFF),
};
*/

const static uint32 ansi_colors32[18*3] =
{
	TO_RGB32(0x00,0x00,0x00),
	TO_RGB32(0xC0,0x00,0x00),
	TO_RGB32(0x00,0xC0,0x00),
	TO_RGB32(0xC0,0xC0,0x00),
	TO_RGB32(0x00,0x00,0xC0),
	TO_RGB32(0xC0,0x00,0xC0),
	TO_RGB32(0x00,0xC0,0xC0),
	TO_RGB32(0xC0,0xC0,0xC0),

	TO_RGB32(0xA0,0x70,0x30),
	TO_RGB32(0xFF,0x00,0x00),
	TO_RGB32(0x00,0xFF,0x00),
	TO_RGB32(0xFF,0xFF,0x00),
	TO_RGB32(0x00,0x00,0xFF),
	TO_RGB32(0xFF,0x00,0xFF),
	TO_RGB32(0x00,0xFF,0xFF),
	TO_RGB32(0xFF,0xF0,0xF0),
	TO_RGB32(0x70,0x50,0x50),
	TO_RGB32(0xFF,0xFF,0xFF),
	TO_RGB32(0xF0,0x60,0x20),

};


void (*more_callback)(void)  = POGO_NULL;

int done_init = 0;

enum { FONT_NORMAL, FONT_ITALIC, FONT_BOLD };

typedef struct
{
	Font *fonts[4];		// Fonts for this window
	uint16 *pixels;		// Top left of pixels

	uint16 width;		// Width of window in pixels
	uint16 height;		// Height of window in pixels
	uint16 stride;		// Width of framebuffer
	uchar current_font;

	char *buffer;		// Contents of window

	char *startptr;		// Location of char at top left corner

	uint32 *offsets;	// screen offset for every char
	char *styles;
	uint16 pos;			// Last character (where to append in buffered window)

	uint16 columns;		// Width of one grid line, only used by fixed windows
	uint16 lines;		// Height of window in lines

	uint16 xstep;
	uint16 lineh;		// Height of a single textline, usually font_height or height / lines

	uint16 type;		// Fixed grid or stream buffer
	int flags;			// Flags like LINEWRAP, AUTOSCROLL
	int x,y;			// Current cursor position for text output

	int topx, topy;

	int fgcol;
	int bgcol;

	int mark_start;		// Start of marked area (for selection or block cursor)
	int mark_stop;		// End of marked area - Both are charpositions

	char *wordbuf;		// Used by word wrapper
	char *wordstyles;
	int wordpos;
	int wordlen;

	int linesout;		// Lines printed (used for output pausing)
	int nextmore;

	int margins[4];
	int linedist;

} Window;

#define SMALL(w) (w->width <= 0 || w->height <= 0 || w->lines <= 0 || w->columns <= 0)

#define MAXWIN 16

static Window *windows[MAXWIN];
static Window *awin;	// Active window

static int console_mark_internal(int start, int stop);
static int mark_area(int start, int stop);

int found_start, found_end;
Window *found_win;
int old_start = -1;
int old_end;

int console_findword(int x, int y, char *dest)
{
	Window *w;
	int i;
	
	found_win = POGO_NULL;
	found_start = found_end = -1;

	if(!done_init)
		return 0;

	for(i=0; i<MAXWIN; i++)
	{
		if((w = windows[i]))
		{
			if((x >= w->topx) && (x < (w->topx + w->width)) &&
				(y >= w->topy) && (y < (w->topy + w->height)))
			{

				//dprintf("%d %d in window %d", x, y, i);

				x -= w->topx;
				y -= w->topy;

				if(w->type == CONWIN_STREAM)
				{
					int j;
					for(j=0; j<w->pos; j++)
					{
						int yw = (w->offsets[j] >> 16) * w->lineh;
						int xw = w->offsets[j] & 0xffff;

						//dprintf("%d %d vs %d %d", x,y,xw,yw);
						if( (yw + w->lineh > y) && (xw > x))
						{
							char *oldd = dest;
							// Found!
							//dprintf("%d %d vs %x %x", xw,yw, w->buffer[j], w->buffer[j+1]);
							char *ptr = &w->buffer[j-1];

							while(*ptr >= 0x30)
								ptr--;
							ptr++;

							j = (int)(ptr - w->buffer);

							while(*ptr >= 0x30)
								*dest++ = *ptr++;
							*dest++ = ' ';
							*dest = 0;

							found_start = j;
							found_end =  (int)(j + strlen(oldd) - 1);
							found_win = w;

							return (int)strlen(oldd);
						}
					}
				}
			}
		}
	}

	return 0;
}

void console_markword(int x, int y)
{
	Window *oldw;
	char dest[64];
	
	if(x >= 0)
	{
		if(console_findword(x, y, dest))
		{
			if(old_start == found_start && old_end == found_end)
				return;
		}
	}
	
	if(old_start > 0)
		mark_area(old_start, old_end);

	old_start = -1;
	if(x >= 0 && found_win && found_start >= 0 && found_end >= 0)
	{
		oldw = awin;
		awin = found_win;
		mark_area(found_start, found_end);
		awin = oldw;
		old_start = found_start;
		old_end = found_end;
	}

}



// For known hardware we initialize the framebuffer pointer correctly
// (Other implementations needs to call console_set_target()
// NOTE that currently we assume a 16bit pixelformat
/*
#if (PLATFORM == PSP)
void *consolePixels = (void *)0x44000000;
int consoleStride = 512;
#elif (PLATFORM == NDS)
void *consolePixels = (void *)0x06800000;
int consoleStride = 256;
#elif (PLATFORM == GBA)
void *consolePixels = (void *)0x06000000;
int consoleStride = 240;
#else*/
Pixel *consolePixels;
int consoleStride;
int consoleHeight;
//#endif

void console_set_target(void *pixels, int stride)
{
	if(!pixels)
	{
		screen_get_target(&consolePixels, &consoleStride, &consoleHeight, POGO_NULL);
	}
	else
	{
		consolePixels = pixels;
		consoleStride = stride;
	}
}

void console_mark_square(int x, int y)
{
	int h = awin->lineh;
	
	//return; // BROKEN
	
	//uint16 *pixels =  &awin->pixels[(x * awin->xstep) + y * awin->stride * awin->lineh];
	
	x = x * awin->xstep + awin->topx;
	y = y * awin->lineh + awin->topy;

	while(h--)
	{
		int w = awin->xstep;
		while(w--)
			*(screen_pixel(x+w, y+h)) ^= 0x7FFF;
		//pixels[w] = pixels[w] ^ 0x7FFF;
		//pixels += awin->stride;
	}
}

static int mark_area(int start, int stop)
{
	int x0, y0, x1, y1;
	if(start == 0 && awin->pos == 0)
		x0 = y0 = x1 = y1 = 0;
	else
	if(awin->type == CONWIN_STREAM)
	{
		 x0 = (awin->offsets[start] & 0xffff) * awin->xstep + awin->topx;
		 y0 = (awin->offsets[start] >> 16) * awin->lineh + awin->topy;
		 x1 = (awin->offsets[stop] & 0xffff) * awin->xstep + awin->topx;;
		 y1 = (awin->offsets[stop] >> 16) * awin->lineh + awin->lineh + awin->topy;

	}
	else
	{
		x0 = (start % awin->columns) * awin->xstep + awin->topx;;
		y0 = (start / awin->columns) * awin->lineh + awin->topy;
		x1 = (stop % awin->columns) * awin->xstep + awin->topx;;
		y1 = (stop / awin->columns) * awin->lineh + awin->lineh + awin->topy;
	}

	if(start == stop)
	{
		int h = awin->lineh;
		//uint16 *pixels =  &awin->pixels[x0 + y0 * awin->stride];
		
		//printf("Marking %d => %d,%d\n", start, x0, y0);

		while(h--)
		{
			//pixels[0] = pixels[0] ^ 0x7FFF;
			//pixels[1] = pixels[1] ^ 0x7FFF;
			//pixels += awin->stride;			
			*(screen_pixel(x0, y0+h)) ^= 0x7FFF;
			*(screen_pixel(x0+1, y0+h)) ^= 0x7FFF;
			
		}
		
		
				
	}
	else
	{
		int h = y1 - y0;
		uint16 *pixels =  &awin->pixels[x0 + y0 * awin->stride];

		while(h--)
		{
			int w = (x1 - x0);
			while(w--)
				*(screen_pixel(x0+w, y0+h)) ^= 0x7FFF;
				//pixels[w] = pixels[w] ^ 0x7FFF;
			//pixels += awin->stride;
		}
	}
	return 0;
}

// This function finds the pixeloffsets for the currently marked area
// and XORs it.
static int console_mark_internal(int start, int stop)
{
	if(start < 0)
		return 0;

	if(SMALL(awin))
		return 0;

	assert(start >= 0 && stop >= 0);
	assert(start <= stop);
	//assert(start <= awin->pos && stop <= awin->pos);

	mark_area(start, stop);

	//awin->mark_start = start;
	//awin->mark_stop = stop;

	return 0;

}

static int console_mark(int start, int stop)
{
	// Remove previous mark if any
	console_mark_internal(awin->mark_start, awin->mark_stop);
	// And mark new if valid
	if(start >= 0)
		console_mark_internal(start, stop);
	
	awin->mark_start = start;
	awin->mark_stop  = stop;

	return 0;
}


static void console_cls(void)
{
	int h = awin->height + (awin->margins[2] + awin->margins[3]);
	int w = awin->width + (awin->margins[0] + awin->margins[1]);
	
	screen_set_target(awin->pixels, 0, 0);
	
	screen_set(awin->topx - awin->margins[0], awin->topy - awin->margins[2], w, h, UINT2PIXEL(awin->bgcol));
	
	/*
	uint16 *p = awin->pixels - awin->margins[0] - (awin->margins[2] * awin->stride);
	uint16 col = UINT2PIXEL(awin->bgcol);
	int h = awin->height + (awin->margins[2] + awin->margins[3]);
	int width = awin->width + (awin->margins[0] + awin->margins[1]);

	//console_mark_internal(awin->mark_start, awin->mark_stop);
	awin->mark_start = -1;
	
	while(h--)
	{
		int i = width;
		while(i--)
			*p++ = col;
		p += (awin->stride - width);	
	}
	
	*/
}

static void console_clear(void)
{
	console_cls();

	if(awin->type == CONWIN_FIXED)
	{
		int i;
		for(i=0; i<(awin->columns * awin->lines); i++)
			awin->buffer[i] = ' ';
	}
	else
		*(awin->buffer) = 0;

	awin->mark_start = -1;

	awin->pos = 0;
	awin->x = awin->y = 0;
}

static void console_gotoxy(int x, int y)
{
	awin->x = x;
	awin->y = y;
}

static void console_setcolor(int fg, int bg)
{
	awin->fgcol = ansi_colors32[fg];
	awin->bgcol = ansi_colors32[bg];
}


static void console_scroll(int y)
{
	int h;
	int modulo = awin->stride - awin->width;
	uint16 bgcol = UINT2PIXEL(awin->bgcol);
	uint16 *dest = &awin->pixels[awin->stride * awin->topy];
	uint16 *src = &awin->pixels[awin->stride * (awin->topy + (y * awin->lineh))];

	if(SMALL(awin))
		return;

	console_mark_internal(awin->mark_start, awin->mark_stop);
	awin->mark_start = -1;

	h = awin->height - (y * awin->lineh);

	if(h <= 0)
	{
		console_clear();
		return;
	}

	screen_scroll(awin->topx, awin->topy, awin->width, awin->height, y * awin->lineh);

	/*
	while(h--)
	{
		w = awin->width;
		while(w--)
			*dest++ = *src++;
		dest += modulo;
		src += modulo;
	}

	h = y * awin->lineh;
	while(h--)
	{
		w = awin->width;
		while(w--)
			*dest++ = bgcol;
		dest += modulo;
	}
	
	*/
	
	if(awin->type == CONWIN_STREAM)
	{
		int i = 0;
		while((int)(awin->offsets[i] >> 16) < y)
			i++;

		awin->pos -= i;

		memmove(awin->buffer, &awin->buffer[i], awin->pos);
		memmove(awin->styles, &awin->styles[i], awin->pos);
		memmove(awin->offsets, &awin->offsets[i], (awin->pos+1) * sizeof(uint32));

		for(i=0; i<awin->pos+1; i++)
			awin->offsets[i] = (awin->offsets[i] - (y<<16));
	}
	else
	{
		memmove(awin->buffer, &awin->buffer[y * awin->columns], awin->columns*(awin->lines-y));
		memset(&awin->buffer[awin->columns*(awin->lines-y)], ' ', awin->columns * y);
	}

	awin->y -= y;

}



static void console_erase(int n)
{
	int clearw,h;

	//uint16 bgcol = UINT2PIXEL(awin->bgcol);
	int oldy = awin->y;

	if(SMALL(awin))
		return;

	if(!n)
		return;

	console_mark_internal(awin->mark_start, awin->mark_stop);
	awin->mark_start = -1;

	if(awin->type == CONWIN_STREAM)
	{
		if(n > awin->pos)
			awin->pos = 0;
		else
			awin->pos -= n;

		awin->x = awin->offsets[awin->pos] & 0xff;
		awin->y = awin->offsets[awin->pos] >> 16;

		//ptr = &awin->pixels[awin->x * awin->xstep + awin->y * awin->lineh * awin->stride];
		h = awin->lineh;
		clearw = awin->width - awin->x;

	}
	else
	{
		while(n--)
		{
			awin->x--;
			if(awin->x < 0)
			{
				awin->x = awin->columns - 1;
				awin->y--;
			}
		}

		//ptr = &awin->pixels[awin->x * awin->xstep + awin->y * awin->lineh * awin->stride];
		h = awin->lineh;
		clearw = awin->width - (awin->x * awin->xstep);
	}
	
	screen_set(awin->x * awin->xstep + awin->topx, awin->y * awin->lineh + awin->topy,
			   clearw, h, UINT2PIXEL(awin->bgcol));
	
	h = (oldy - awin->y) * awin->lineh;

	if(h)
	{
		screen_set(awin->topx, (awin->y + 1) * awin->lineh + awin->topy,
				   awin->width, h, UINT2PIXEL(awin->bgcol));
	}
	
	return; // BROKEN

	// Clear area
	/*
	while(h--)
	{
		w = clearw;
		while(w--)
			*ptr++ = bgcol;
		ptr += (awin->stride - clearw);
	}

	ptr -= (awin->width - clearw);

	h = (oldy - awin->y) * awin->lineh;

	while(h--)
	{
		w = awin->width;
		while(w--)
			*ptr++ = bgcol;
		ptr += (awin->stride - awin->width);
	}
	*/
}

static int console_getcharcount()
{
	if(awin->type == CONWIN_FIXED)
		return awin->x + awin->y * awin->columns;
	else
		return awin->pos;
}


static void console_putchar(unsigned char c)
{
	int w = 0;
	Font *font = awin->fonts[awin->current_font];

	if(!c)
		font_setcolor(awin->fgcol, awin->bgcol, font);

	if(SMALL(awin))
		return;

	// If there are chars to print we remove mark first
	if(awin->wordpos || c)
	{
		console_mark_internal(awin->mark_start, awin->mark_stop);
		awin->mark_start = -1;
	}

	if(c == 12)
	{
		console_clear();
		return;
	}

	screen_set_target(awin->pixels, 0, 0);

	if(awin->type == CONWIN_FIXED)
	{
		switch(c)
		{
		case 13:
		case 10:
			awin->x = 0;
			awin->y++;
			awin->linesout++;
			break;
		case 9:
			break;
		default:
			break;
		}

		if((awin->x + 1) > awin->columns)
		{
			return;
			//awin->x = 0;
			//awin->y++;
			/*if(0)//awin->flags & CM_LINEWRAP)
			{

				awin->y++;
				awin->linesout++;
			}
			else
				return;*/
		}

		// TODO: Scrolling of fixed windows ?
		//while(awin->y >= awin->lines)
		//{
		//	console_scroll(1);
		//}

		if(awin->y >= awin->lines)
		{
			return;
		}

		if(c >= 0x20)
		{
			awin->buffer[awin->x + awin->y * awin->columns] = c;
			font_putchar(font, c, awin->topx + awin->x * awin->xstep, awin->topy + awin->y * awin->lineh);
			awin->x++;
		}
	}
	else
	{
		int i;
		//int wordwrap = 0;
		if(c && (awin->flags & CM_WORDWRAP))
		{
			awin->wordbuf[awin->wordpos] = c;
			awin->wordstyles[awin->wordpos] = awin->current_font;
			//awin->offsets[awin->endpos] = awin->x | (awin->y << 16);
			awin->wordpos++;
			w = font_putchar(font, c, -1, -1);
			awin->wordlen += w;

			if(c != ' ' && c != 13 && c != 10)
				return;

			if(awin->wordlen + awin->x > awin->width)
			{
				awin->x = 0;
				awin->y++;
				awin->linesout++;

			}
		}
		else
		if(c)
		{
			awin->wordbuf[awin->wordpos++] = c;
			awin->wordstyles[awin->wordpos] = awin->current_font;
		}

		if(more_callback && (awin->linesout == awin->nextmore))
		{
			int w;
			Window *win;
			awin->nextmore = awin->linesout + awin->lines - 1;
			while(awin->y >= awin->lines)
				console_scroll(1);
			
			w = font_text(awin->fonts[0], "[MORE]", awin->topx, awin->topy + (awin->lines-1) * awin->lineh);
			win = awin;
			more_callback();
			awin = win;
			font_setcolor(awin->fgcol, awin->bgcol, font);
			screen_set_target(awin->pixels, 0, 0);
			screen_set(awin->topx, awin->topy + (awin->lines-1) * awin->lineh, w, awin->fonts[0]->height, screen_color(0));
		}


		while(awin->y >= awin->lines)
			console_scroll(1);


		//while(awin->pos < awin->endpos)
		for(i=0; i<awin->wordpos; i++)
		{
			c = awin->wordbuf[i];
			font = awin->fonts[(int)awin->wordstyles[i]];
			awin->buffer[awin->pos] = c;

			awin->offsets[awin->pos] = awin->x | (awin->y << 16);
			awin->styles[awin->pos] = awin->wordstyles[i];
			awin->pos++;


			switch(c)
			{
			case 13:
			case 10:
				w = 0;
				awin->x = 0;
				awin->y++;
				awin->linesout++;
				break;
			case 9:
				w = 0;
				break;
			default:
				w = font_putchar(font, c, -1, -1);
				break;
			}


			awin->offsets[awin->pos] = (awin->x + w) | (awin->y << 16);
			//awin->styles[awin->pos] = awin->current_font;

			if(w + awin->x > awin->width)
			{
				//if(awin->flags & CM_LINEWRAP)
				//{
					awin->x = 0;
					awin->y++;
					awin->linesout++;
				//}
				//else
				//	return;
			}

			while(awin->y >= awin->lines)
				console_scroll(1);

			if(w)
			{
				font_putchar(font, c, awin->topx + awin->x, awin->topy + awin->y * awin->lineh);
				awin->x += w;
			}
		}
		awin->wordpos = 0;
		awin->wordlen = 0;

		if(more_callback && (awin->linesout == awin->nextmore))
		{
			int w;
			awin->nextmore = awin->linesout + awin->lines - 1;
			w = font_text(awin->fonts[0], "[MORE]", awin->topx, awin->topy + (awin->lines-1) * awin->lineh);
			more_callback();
			font_setcolor(awin->fgcol, awin->bgcol, font);
			screen_set_target(awin->pixels, 0, 0);
			screen_set(awin->topx,  awin->topy + (awin->lines-1) * awin->lineh, w, awin->fonts[0]->height, screen_color(0));
		}

	}

	//if(!c)
	//{
	//}

}

static void console_setfont(int num)
{
	//awin->fonts[num] = font;
	awin->current_font = num;
}

static void window_calc(Window *win)
{
	int i;
	int fw = 0;
	int fh = 0;

	// Find dimensions of largest font
	for(i=0; i<4; i++)
	{
		if(win->fonts[i])
		{
			if(win->fonts[i]->charwidth + win->fonts[i]->spacing > fw)
				fw = win->fonts[i]->charwidth + win->fonts[i]->spacing;
			if(win->fonts[i]->height + win->linedist > fh)
				fh = win->fonts[i]->height + win->linedist;
		}
	}

	win->columns = win->width / fw;
	win->lines = win->height / fh;
	
	win->lineh = fh;//win->fonts[0]->height;

	if(win->type == CONWIN_FIXED)
		win->xstep = fw;// win->width / win->columns;
	else
		win->xstep = 1;
}

void console_window_close(int id)
{
	Window *save = awin;
	if(awin == windows[id])
		save = POGO_NULL;
	awin = windows[id];
	console_clear();
	
	if(awin->buffer)
		free(awin->buffer);
	if(awin->offsets)
		free(awin->offsets);
	if(awin->styles)
		free(awin->styles);
	if(awin->wordbuf)
		free(awin->wordbuf);
	if(awin->wordstyles)
		free(awin->wordstyles);

	free(awin);
	windows[id] = POGO_NULL;
	awin = save;
}

static int console_window_create(int x, int y, int w, int h, int stride, Font **fonts, int type)
{
	int i;
	Window *win;
	int id = -1;
	void *pixels;


	//int fw = 0;
	//int fh = 0;

	for(i=0; i<MAXWIN; i++)
		if(!windows[i])
		{
			id = i;
			break;
		}

	if(id < 0)
		return id;


	win = (Window *)malloc(sizeof(Window));
	win->linedist = 0;
	win->margins[0] = win->margins[1] = 0;
	win->margins[2] = win->margins[3] = 0;

	if(type == CONWIN_STREAM)
	{
		win->linedist = 0;
		win->margins[0] = win->margins[1] = 4;
		win->margins[2] = 4;
		win->margins[3] = 0;
	}

	if((win->margins[0] + win->margins[1]) < w)
		w -= (win->margins[0] + win->margins[1]);
	else
	{
		win->margins[0] = win->margins[1] = 0;
	}

	if((win->margins[2] + win->margins[3]) < h)
		h -= (win->margins[2] + win->margins[3]);
	else
	{
		win->margins[2] = win->margins[3] = 0;
	}

	//pixels = (uint16 *)((uchar *)consolePixels + (x+win->margins[0])*2 + consoleStride * 2 * (y+win->margins[2]));	
	pixels = consolePixels;
	
	
	win->width = w;
	win->height = h;
	win->stride = stride;

	win->topx = x + win->margins[0];
	win->topy = y + win->margins[2];

	for(i=0; i<4; i++)
		win->fonts[i] = fonts[i];

	win->current_font = 0;

	//win->startpos = win->buffer;
	//win->endpos = win->buffer;
	win->pos = 0;

	win->pixels = (uint16*)pixels;//(uint16 *)(0x06800000 + x*2 + 512 * y);

	win->x = 0;
	win->y = 0;

	win->type = type; //CONWIN_FIXED;
	win->flags = CM_LINEWRAP | CM_WORDWRAP;

	win->mark_start = -1;

	windows[id] = win;

	win->linesout = 0;

	window_calc(win);
	win->nextmore = win->lines - 1;


	// TODO: Grow buffer as needed instead of large initial size

	if(type == CONWIN_STREAM)
	{
		win->buffer = (char *)malloc(1000 * 100);
		win->offsets = (uint32 *)malloc(1000 * 100 * sizeof(uint32));
		win->styles = (char *)malloc(1000 * 100 * 12);
	}
	else
	{
		win->buffer = malloc(win->lines * win->columns);
		win->offsets = POGO_NULL;
		win->styles = POGO_NULL;
	}

	win->wordbuf = (char *)malloc(256);
	win->wordstyles = (char *)malloc(256);
	win->wordpos = win->wordlen = 0;
	awin = win;

	console_setcolor(15, 0);
	console_clear();

	return id;

}

int console_write(int fd, const void *dest, int size)
{
	int i;
	Font *font = awin->fonts[awin->current_font];
	uchar *s = (uchar *)dest;
	font_setcolor(awin->fgcol, awin->bgcol, font);
	for(i=0; i<size; i++)
		console_putchar(s[i]);

	return size;
}

int console_read(int fd, void *dest, int size)
{
	return 0;
}

int console_lseek(int fd, int offset, int origin)
{
	return -1;
}


static void console_move(int x, int y, int neww, int newh)
{
	int oldc, oldl;
	Font *font = awin->fonts[awin->current_font];
	font_setcolor(awin->fgcol, awin->bgcol, font);

	//if(awin->mark_start >= 0)
	//	console_mark_internal(awin->mark_start, awin->mark_stop);
	awin->mark_start = -1;


	if(awin->type == CONWIN_STREAM)
	{
		awin->margins[0] = awin->margins[1] = 4;
		awin->margins[2] = 4;
		awin->margins[3] = 0;
	}

	if((awin->margins[0] + awin->margins[1]) < neww)
		neww -= (awin->margins[0] + awin->margins[1]);
	else
	{
		awin->margins[0] = awin->margins[1] = 0;
	}

	if((awin->margins[2] + awin->margins[3]) < newh)
		newh -= (awin->margins[2] + awin->margins[3]);
	else
	{
		awin->margins[2] = awin->margins[3] = 0;
	}
	awin->topx = x + awin->margins[0];
	awin->topy = y + awin->margins[2];

	
	//if(x >= 0 && y >= 0)
	//	awin->pixels = (uint16 *)((uchar *)consolePixels + (x+awin->margins[0])*2 + consoleStride * 2 * (y+awin->margins[2]));

	//neww -= (awin->margins[0] + awin->margins[1]);
	//newh -= (awin->margins[2] + awin->margins[3]);


	awin->x = awin->y = 0;
	awin->stride = consoleStride;

	screen_set_target(awin->pixels, 0, 0);

	if(neww >=0)
		awin->width = neww;

	if(newh >=0)
		awin->height = newh;

	oldc = awin->columns;
	oldl = awin->lines;

	window_calc(awin);
	
	if(awin->type == CONWIN_FIXED)
		awin->buffer = realloc(awin->buffer, awin->lines * awin->columns);

	awin->nextmore -= (oldl - awin->lines);


	console_cls();
/*
	{
		uint16 *p = awin->pixels;
		uint16 col = UINT2PIXEL(awin->bgcol);
		int h = awin->height;
		while(h--)
		{
			int i = awin->width;
			while(i--)
				*p++ = col;
			p += (awin->stride - awin->width);
		}
	}
*/
	if(awin->type == CONWIN_FIXED)
	{
		int x,y;

		for(y=0; y<awin->lines; y++)
			for(x=0; x<awin->columns; x++)
			{
				//if(x < oldc && y < oldl)
				//	console_putchar(awin->buffer[x + y * oldc]);
				//else
					console_putchar(' ');
			}

	}
	else
	{
		int oldpos, oldnextmore, oldlinesout;
		int i;
		//char c;
		//int w = 0;
		//Font *font = awin->fonts[awin->current_font];


		oldpos = awin->pos;
		awin->pos = 0;

		//awin->linesout = 0;
		//awin->nextmore = awin->lines - 1;
		oldnextmore = awin->nextmore;
		oldlinesout = awin->linesout;
		awin->nextmore = awin->linesout + 1000;

		for(i=0; i<oldpos; i++)
		{
			if(awin->styles[i] != awin->current_font)
				awin->current_font = awin->styles[i];
			console_putchar(awin->buffer[i]);
		}

		awin->nextmore = oldnextmore;
		awin->linesout = oldlinesout;
	}

	//if(awin->mark_start >= 0)
	//	console_mark_internal(awin->mark_start, awin->mark_stop);
}

int console_ioctl_v(int fd, int req, va_list vl)
{
	char *s;
	int c;
	
	if(awin)
		screen_set_target(awin->pixels, 0, 0);

	switch(req) {
	case CC_GETTEXTWIDTH:
		s = va_arg(vl, char*);
		return font_text(awin->fonts[awin->current_font], s, -1, -1);
	case CC_GETCHARWIDTH:
		c = va_arg(vl, int);
		return font_putchar(awin->fonts[awin->current_font], c, -1, -1);
	case CC_SETFONT:
		console_setfont(va_arg(vl, int));
		return 0;
	case CC_GETFONT:
		return awin->current_font;//(int)awin->fonts[va_arg(vl, int)];
	case CC_GETX:
		return awin->x;
	case CC_GETY:
		return awin->y;
	case CC_GETCHAR:
		return awin->buffer[awin->x + awin->y * awin->columns];
	case CC_LINESOUT:
		return awin->linesout;
		break;
	case CC_GOTOXY:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			FLUSH();
			console_gotoxy(x, y);
		}
		return 1;
	case CC_MARK:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			FLUSH();
			console_mark(x, y);
		}
		return 1;
	case CC_ERASE:
		FLUSH();
		console_erase(va_arg(vl, int));
		return 1;
	case CC_SETCOLOR:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			FLUSH();
			console_setcolor(x, y);
		}
		return 1;
	case CC_SETRGB:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			FLUSH();
			awin->fgcol = x;
			awin->bgcol = y;
		}
		return 1;
	case CC_GETXY:
	case CC_GETPOS:
		FLUSH();
		return console_getcharcount();
	case CC_CLEAR:
		FLUSH();
		console_clear();
		return 1;
	case CC_GETHEIGHT:
		return awin->lines;
	case CC_GETWIDTH:
		return awin->columns;
	case CC_SCROLL:
		FLUSH();
		console_scroll(va_arg(vl, int));
		return 1;
	case CC_FLUSH:
		console_putchar(0);
		//display_flush();
		return 1;
	case CC_SETWIN:
		{
			int w = va_arg(vl, int);
			if(w < 0)
			{
				int *p = 0;
				*p = 1;
			}
						
			if(awin != windows[w])
			{
				if(awin)
					FLUSH();
				awin = windows[w];
			}
		}
		return 1;
	case CC_OPENWIN:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			int w = va_arg(vl, int);
			int h = va_arg(vl, int);
			Font **fonts = (va_arg(vl, Font**));
			int type = va_arg(vl, int);
			return console_window_create(x, y, w, h, consoleStride, fonts, type);
		}
	case CC_CLOSEWIN:
		console_window_close(va_arg(vl, int));
		return 1;
	case CC_RESIZEWIN:
		{
			int w = va_arg(vl, int);
			int h = va_arg(vl, int);
			FLUSH();
			console_move(-1, -1, w, h);
		}
		return 1;
	case CC_MOVEWIN:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			int w = va_arg(vl, int);
			int h = va_arg(vl, int);
			FLUSH();
			console_move(x, y, w, h);
		}
		return 1;
	case CC_SETMORECALLBACK:
		more_callback = (void (*)(void))va_arg(vl, void*);
		break;
	case CC_RESETMORE:
		awin->nextmore = awin->linesout + awin->lines - 1;
		break;
	}
	return -1;
}

int console_ioctl(int fd, int request, ...)
{
	int rc = -1;
	va_list vl;
	va_start(vl, request);
	rc = console_ioctl_v(fd, request, vl);
	va_end(vl);
	return rc;
}

int console_open(const char *name, int mode)
{
	Font *font[4];
	int w, h;
	
	screen_get_target(POGO_NULL, &w, &h, POGO_NULL);

	if(strlen(name) != 0)
		return -1;

    // Open a default window
	font[0] = font[1] = font[2] = font[3] = font_getdefault();
	console_ioctl(0, CC_OPENWIN, 0, 0, w, h, font, CONWIN_FIXED);
	console_ioctl(0, CC_SETCOLOR, CCOL_WHITE, CCOL_BLACK);
	return 0;

}


static PogoDevice condev;


void console_init()
{
	int i;
	for(i=0; i<MAXWIN; i++)
		windows[i] = POGO_NULL;

	screen_get_target(&consolePixels, &consoleStride, &consoleHeight, POGO_NULL);
	
	done_init = 1;

	more_callback = POGO_NULL;

	memset(&condev, 0, sizeof(condev));
	condev.open = console_open;
	condev.write = console_write;
	condev.ioctl = console_ioctl_v;
	condev.read = console_read;
	condev.lseek = console_lseek;

	pogo_device_register(&condev, "/dev/console", POGO_NULL, 1);

}

