/*
 * Core memory handling functions (memory.c)
 * Part of libpogo, a c-library replacent for GBA
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 * Functions for allocating and freeing memory in WRAM
 *
 * New System:
 * "MemInfo" structures keep track of allocated areas and
 * are stored as a reverse-growing vector at the end of
 * memory in no particular order (allocation just gets the
 * next availble info-structure.
 *
 *
 **/

#undef MALLOC_DEBUG
#include "pogo/stdcore.h"

void *mdbgptr;


//typedef unsigned int uint32;
//#include <stdio.h>
//#include <stdlib.h>

#ifndef MEM_STRUCTS

typedef struct _MemInfo
{
	uint32 size;
	uint32 *data;
} MemInfo;


typedef struct
{
	uint32 *base;
	int size;
	MemInfo *bottom;
	MemInfo *top;

} MemArea;

#endif

void init_area(MemArea *area, uint32 *memory, int size)
{
	while(size < 32);

	area->base = memory;
	area->size = size;
	area->bottom = area->top = (MemInfo *)&area->base[size-2];

	area->top->data = (uint32*)(area->top-1);
	area->top->size = 0;
	area->top--;
}

// Bottom points to FIRST meminfo
// Top points to next unsused meminfo


/* Alloc 32bit words */
uint32 *memory_alloc(MemArea *area, uint32 *start, int alloc_size)
{
	int free_before;
	MemInfo saved_info;
	uint32 *ptr;

	uint32 *last_end = area->base;
	MemInfo *info = area->bottom;

	while(info > area->top)
	{
		if(start)
			last_end = start;

		free_before = info->data - last_end;

		//if(start)
		//	pogo_fprintf(2, "Free before %p : %d\n", info->data, free_before);

		if(free_before >= alloc_size)
		{

			saved_info = *info;
			info->data = last_end;
			info->size = alloc_size;
			ptr = last_end;

			while(info > area->top)
			{
				MemInfo tmp_info = saved_info;
				info--;
				saved_info = *info;
				*info = tmp_info;
			}
			area->top--;
			area->top[1].data -= 2;

			//pogo_fprintf(2, "alloc %d => %p\n", alloc_size, ptr);

			return ptr;

		}

		last_end = &info->data[info->size];

		info--;
	}

	return POGO_NULL;
}
/*
void print_stack(MemArea *area)
{
	char tmp[32];
	MemInfo *info = area->bottom;
	sprintf(tmp, "---------------------\n");
	dprint(tmp);
	while(info > area->top)
	{
		sprintf(tmp, "%p %d\n", info->data, info->size);
		dprint(tmp);
		info--;
	}
	sprintf(tmp, "---------------------\n");
	dprint(tmp);
}
*/

uint32 *memory_free(MemArea *area, uint32 *start)
{
	MemInfo *info = area->bottom;
	while(info > area->top)
	{
		if(info->data == start)
		{
			while(info > area->top)
			{
				info--;
				info[1] = info[0];
			}
			area->top++;
			area->top[1].data += 2;
			return start;
		}
		info--;
	}
	return POGO_NULL;
}

uint32 *memory_realloc(MemArea *area, uint32 *start, int size)
{
	MemInfo *info = area->bottom;

	if(!start)
		return memory_alloc(area, POGO_NULL, size);

	while((info-1) > area->top)
	{
		if(info->data == start)
		{
			uint32 *ptr;
			int free_after = info[-1].data - &info->data[info->size];

			// If new size is smaller just shrink and return
			if((size <= info->size) || (size <= (info->size + free_after)))
			{
				info->size = size;
				return info->data;
			}

			// Allocate new area and return
			ptr = memory_alloc(area, POGO_NULL, size);
			memcpy(ptr, info->data, info->size * 4);
			memory_free(area, info->data);
			return ptr;
		}

		info--;
	}
	return POGO_NULL;
}
int memory_avail(MemArea *area)
{

	int total = 0;
	uint32 *last_end = area->base;
	MemInfo *info = area->bottom;

	while(info > area->top)
	{
		total += (info->data - last_end);
		info--;
	}

	return total;

}


///////////// PART TWO //////////////



static int current_area = -1;
static MemArea mem_areas[4];

/* Exported by linkscript as start of available memory (after any WRAM variables) */
extern uint32 __eheap_start;
extern uint32 __iheap_start;
extern uint32 __eheap_end;
extern uint32 __iheap_end;

uint32 fake_heap_end;

void memory_init(void)
{

#ifdef SOMETHING_XXX

#if (defined GBA)
	init_area(&mem_areas[0], &__eheap_start, (0x02400000 - (int)&__eheap_start) / 4);
	init_area(&mem_areas[1], &__iheap_start, (0x03007E00 - (int)&__iheap_start) / 4);
	init_area(&mem_areas[2], (uint32*)0x06000000, 96*256);
	init_area(&mem_areas[3], (uint32*)0x08400000, 4*1024*1024); // EZ3 RAM
#elif (defined DS)
	//init_area(&mem_areas[0], &__eheap_start, (0x02400000 - (int)&__eheap_start) / 4);
	init_area(&mem_areas[0], (void *)0x02080000, (0x02380000 - 0x02080000) / 4);
	init_area(&mem_areas[1], &__iheap_start, (0x03007E00 - (int)&__iheap_start) / 4);
	//init_area(&mem_areas[0], &__eheap_start, (&__eheap_end - &__eheap_start) / 4);
	//init_area(&mem_areas[1], &__iheap_start, (&__iheap_start - (int)&__iheap_start) / 4);
#elif (defined PSP)
	init_area(&mem_areas[0], (void *)0x08980000, 5*1024*1024);
#endif

#endif

	current_area = 0;
}

void memory_setarea(int area)
{
	if(current_area == -1)
		memory_init();
	current_area = area;
}

#ifdef SOMETHING_XXX

void *calloc(int x, int y)
{
	void  *dst = malloc(x*y);

	if(!dst)
		return POGO_NULL;

	//if(y == 1)
	//	memset8(dst, 0, x*y);
	//else
	memset(dst, 0, x*y);

	return dst;
}

void *_malloc(int x)
{
	void *p = (void *)memory_alloc(&mem_areas[current_area], POGO_NULL, (x+3)/4);
	return p;
};

void *_malloc_fixed(void *ptr, int x)
{
	void *p = (void *)memory_alloc(&mem_areas[current_area], ptr, (x+3)/4);
	return p;
};

void *_realloc(void *p, int l)
{
	return (void *)memory_realloc(&mem_areas[current_area], (uint32*)p, (l+3)/4);
}

void _free(void *x)
{
	memory_free(&mem_areas[current_area], (uint32*)x);
}

void *malloc(int x)
{
	void *p = (void *)memory_alloc(&mem_areas[current_area], POGO_NULL, (x+3)/4);
	return p;
};

void *malloc_fixed(void *ptr, int x)
{
	void *p = (void *)memory_alloc(&mem_areas[current_area], ptr, (x+3)/4);
	return p;
};

void *realloc(void *p, int l)
{
	return (void *)memory_realloc(&mem_areas[current_area], (uint32*)p, (l+3)/4);
}

void free(void *x)
{
	memory_free(&mem_areas[current_area], (uint32*)x);
}


int avail(void)
{
	return memory_avail(&mem_areas[current_area]);
}

#endif
